/*
    K8400 firmware modification based on original Version provided by Vellemann.

 Copyright (C) 2017 Holger Höhnemann

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#ifndef PARAMETER_H_
#define PARAMETER_H_

#include "MarlinHOH.h"
#include "temperature.h"

#define EEPROM_OFFSET 100

class parameter
{
public:
	parameter();
	virtual ~parameter();



	//
	// parameters stored in EEPROM
	//
	uint16_t 	checksum_bytes;
	float	axis_steps_per_unit[NUM_AXIS];
	float	max_feedrate_mm_s[NUM_AXIS]; // set the max speeds in mm/sec

	float 	min_feedrate_mm_s;
	float 	acceleration;         // Normal acceleration mm/s^2  THIS IS THE DEFAULT ACCELERATION for all moves. M204 SXXXX
	float 	retract_acceleration; //  mm/s^2   filament pull-pack and push-forward  while standing still in the other axis M204 TXXXX
	float 	travel_acceleration; // Travel acceleration mm/s^2  DEFAULT ACCELERATION for all NON printing moves. M204 MXXXX

	float  max_jerk[NUM_AXIS]; //speed than can be stopped at once, if i understand correctly.
	//	float 	max_z_jerk;
	//	float 	max_e_jerk;
	float 	min_travel_feedrate_mm_s;
	float 	add_homeing[NUM_AXIS-1];



#ifdef PID_PARAMS_PER_EXTRUDER
	float Kp[EXTRUDERS] ;
	float Ki[EXTRUDERS] ;
	float Kd[EXTRUDERS] ;
#ifdef PID_ADD_EXTRUSION_RATE
	float Kc[EXTRUDERS] ;
	float Kf[EXTRUDERS] ;
#endif // PID_ADD_EXTRUSION_RATE
#else //PID_PARAMS_PER_EXTRUDER
	float Kp = DEFAULT_Kp;
	float Ki = (DEFAULT_Ki) * (PID_dT);
	float Kd = (DEFAULT_Kd) / (PID_dT);
#ifdef PID_ADD_EXTRUSION_RATE
	float Kc = DEFAULT_Kc;
	float Kf = DEFAULT_KF;
#endif // PID_ADD_EXTRUSION_RATE
#endif // PID_PARAMS_PER_EXTRUDER


	// Extruder offset
#if EXTRUDERS > 1
	float extruder_offset[NUM_EXTRUDER_OFFSETS][EXTRUDERS];
#endif

	float bedKp ;
	float bedKi ;
	float bedKd ;


#ifdef PRESSURE_ADVANCE
	float advance_factor[EXTRUDERS];							// the pressure advance factor for the extruders
	int advance_delay[EXTRUDERS];								// the ramp up delay
#endif

	int plaPreheatHotendTemp;
	int plaPreheatHPBTemp;
	int plaPreheatFanSpeed;

	int absPreheatHotendTemp;
	int absPreheatHPBTemp;
	int absPreheatFanSpeed;

	unsigned long max_acceleration_units_per_sq_second[NUM_AXIS];
	unsigned long minsegmenttime;
	unsigned long axis_steps_per_sqr_second[NUM_AXIS];

	void Load_Failsave();
	void Load_EEPROM();
	void Store_EEPROM();

	void Print_Settings();

	void Reset_AccelerationRates();

	// get and set scaled parameters
	void Set_Kp (float value, int e);
	void Set_Ki (float value, int e);
	void Set_Kd (float value, int e);
	void Set_Kc (float value, int e);
	void Set_Kf (float value, int e);
	float Get_Kp (int e);
	float Get_Ki (int e);
	float Get_Kd (int e);
	float Get_Kc (int e);
	float Get_Kf (int e);
	// raw values to be used in PID routine.
	float kp (int e){return (e<0) ? bedKp : Kp[e];};
	float ki (int e){return (e<0) ? bedKi : Ki[e];};
	float kd (int e){return (e<0) ? bedKd : Kd[e];};
	float kc (int e){return (e<0) ? 0.0 : Kc[e];};
	float kf (int e){return (e<0) ? 0.0 : Kf[e];};
private:

	void _EEPROM_writeData(int &pos, uint8_t* value, uint8_t size);
	void _EEPROM_readData(int &pos, uint8_t* value, uint8_t size);
	uint16_t Generate_Check_Sum();
	void Write_Checksum_Bytes();
	bool Checksum_Error();

};

extern class parameter config;


//#define PID_PARAM(param, e) config.param[e] // use macro to point to array value


#endif /* PARAMETER_H_ */
