/*
    K8400 firmware modification based on original Version provided by Vellemann.

 Copyright (C) 2017 Holger Höhnemann

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
#ifndef stepper_h
#define stepper_h 

//#include <planner_old.h>
#include "planner.h"

#if EXTRUDERS > 2
  #define WRITE_E_STEP(v) { if(current_block->active_extruder == 2) { WRITE(E2_STEP_PIN, v); } else { if(current_block->active_extruder == 1) { WRITE(E1_STEP_PIN, v); } else { WRITE(E0_STEP_PIN, v); }}}
  #define NORM_E_DIR() { if(current_block->active_extruder == 2) { WRITE(E2_DIR_PIN, !INVERT_E2_DIR); } else { if(current_block->active_extruder == 1) { WRITE(E1_DIR_PIN, !INVERT_E1_DIR); } else { WRITE(E0_DIR_PIN, !INVERT_E0_DIR); }}}
  #define REV_E_DIR() { if(current_block->active_extruder == 2) { WRITE(E2_DIR_PIN, INVERT_E2_DIR); } else { if(current_block->active_extruder == 1) { WRITE(E1_DIR_PIN, INVERT_E1_DIR); } else { WRITE(E0_DIR_PIN, INVERT_E0_DIR); }}}
#elif EXTRUDERS > 1
  #ifndef DUAL_X_CARRIAGE
    #define WRITE_E_STEP(v) { if(current_block->active_extruder == 1) { WRITE(E1_STEP_PIN, v); } else { WRITE(E0_STEP_PIN, v); }}
    #define NORM_E_DIR() { if(current_block->active_extruder == 1) { WRITE(E1_DIR_PIN, !INVERT_E1_DIR); } else { WRITE(E0_DIR_PIN, !INVERT_E0_DIR); }}
    #define REV_E_DIR() { if(current_block->active_extruder == 1) { WRITE(E1_DIR_PIN, INVERT_E1_DIR); } else { WRITE(E0_DIR_PIN, INVERT_E0_DIR); }}
  #else
    extern bool extruder_duplication_enabled;
    #define WRITE_E_STEP(v) { if(extruder_duplication_enabled) { WRITE(E0_STEP_PIN, v); WRITE(E1_STEP_PIN, v); } else if(current_block->active_extruder == 1) { WRITE(E1_STEP_PIN, v); } else { WRITE(E0_STEP_PIN, v); }}
    #define NORM_E_DIR() { if(extruder_duplication_enabled) { WRITE(E0_DIR_PIN, !INVERT_E0_DIR); WRITE(E1_DIR_PIN, !INVERT_E1_DIR); } else if(current_block->active_extruder == 1) { WRITE(E1_DIR_PIN, !INVERT_E1_DIR); } else { WRITE(E0_DIR_PIN, !INVERT_E0_DIR); }}
    #define REV_E_DIR() { if(extruder_duplication_enabled) { WRITE(E0_DIR_PIN, INVERT_E0_DIR); WRITE(E1_DIR_PIN, INVERT_E1_DIR); } else if(current_block->active_extruder == 1) { WRITE(E1_DIR_PIN, INVERT_E1_DIR); } else { WRITE(E0_DIR_PIN, INVERT_E0_DIR); }}
  #endif  
#else
  #define WRITE_E_STEP(v) WRITE(E0_STEP_PIN, v)
  #define NORM_E_DIR() WRITE(E0_DIR_PIN, !INVERT_E0_DIR)
  #define REV_E_DIR() WRITE(E0_DIR_PIN, INVERT_E0_DIR)
#endif

#ifdef ABORT_ON_ENDSTOP_HIT_FEATURE_ENABLED
extern bool abort_on_endstop_hit;
#endif

ISR (TIMER3_COMPA_vect);

ISR (TIMER5_COMPA_vect);

class stepper
{
	friend void __vector_32();
	friend void __vector_47();

public:
	stepper();
	virtual ~stepper();



	void Initialize();
	void Enable_Endstops(bool check); // Enable/disable endstop checking

	// The stepper subsystem goes to sleep when it runs out of things to execute. Call this
	// to notify the subsystem that it is time to go to work.
	void Wake_Up();

	void QuickStop();
	void FinishAndDisable();
	void Endstops_hit_on_Purpose();
	void CheckHitEndstops();
	void Synchronize();  // Block until all buffered steps are executed

	// Set current position in steps
	void Set_Position(const long &x, const long &y, const long &z, const long &e);
	void Set_E_Position(const long &e);
	void Set_Position (const AxisEnum axus, const long & e);

	// Get current position in steps
	long Get_Position(AxisEnum axis);

//	static void isr();

	bool check_endstops = true;

// subroutines for interrupt
//	FORCE_INLINE void trapezoid_generator_reset();
	FORCE_INLINE void Reset_trapezoid_generator(); // for timer 3
private:

	void  digipot_init(); //Initialize Digipot Motor Current
	void  microstep_init(); //Initialize Microstepping Pins

	inline void isr_TIMER3_COMPA_vect();
	inline void isr_TIMER5_COMPA_vect();
	//variables for timer4
	volatile int steps;
	inline unsigned short calc_timer(unsigned short step_rate) ;
	// Variables used by The Stepper Driver Interrupt
	// variables for timer3
	unsigned char out_bits;        // The next stepping-bits to be output
	long acceleration_time, deceleration_time;
	move_block *current_block = NULL;  // A pointer to the block currently being traced
	long counter_x,       // Counter variables for the bresenham line tracer
	            counter_y,
	            counter_z,
	            counter_e;
	signed char count_direction[NUM_AXIS] = { 1, 1, 1, 1};
	unsigned short acc_step_rate; // needed for deccelaration start point
	unsigned long step_events_completed; // The number of step events executed in the current block
	char step_loops;
	unsigned short OCR3A_nominal ;
	unsigned short step_loops_nominal;

	volatile bool endstop_x_hit=false;
	volatile bool endstop_y_hit=false;
	volatile bool endstop_z_hit=false;



};


#ifdef ENABLE_AUTO_BED_LEVELING
// Get current position in mm
//float st_get_position_mm(uint8_t axis);
#endif  //ENABLE_AUTO_BED_LEVELING



  
//void checkHitEndstops(); //call from somewhere to create an serial error message with the locations the endstops where hit, in case they were triggered
//void endstops_hit_on_purpose(); //avoid creation of the message, i.e. after homing and before a routine call of checkHitEndstops();

// void enable_endstops(bool check); // Enable/disable endstop checking

//void checkStepperErrors(); //Print errors detected by the stepper

//void finishAndDisableSteppers();

//extern block_t *current_block;  // A pointer to the block currently being traced

//void quickStop();

void digitalPotWrite(int address, int value);
void microstep_ms(uint8_t driver, int8_t ms1, int8_t ms2);
void microstep_mode(uint8_t driver, uint8_t stepping);
void digipot_init();
void digipot_current(uint8_t driver, int current);
void microstep_init();
void microstep_readings();

#ifdef BABYSTEPPING
  void babystep(const uint8_t axis,const bool direction); // perform a short step with a single stepper motor, outside of any convention
#endif
     
extern  class stepper motor;

#endif
