/*
    K8400 firmware modification based on original Version provided by Vellemann.

 Copyright (C) 2017 Holger Höhnemann

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
#include "stopwatch.h"
#include "MarlinHOH.h"

Stopwatch::Stopwatch()
{
  reset();
}

void Stopwatch::stop() {
  #if defined DEBUG_STOPWATCH
    debug(PSTR("stop"));
  #endif

  if (!isRunning()) return;

  status = STPWTCH_STOPPED;
  stopTimestamp = millis();
}

void Stopwatch::pause() {
  #if defined(DEBUG_STOPWATCH)
    debug(PSTR("pause"));
  #endif

  if (!isRunning()) return;

  this->status = STPWTCH_PAUSED;
  this->stopTimestamp = millis();

}

void Stopwatch::start() {
  #if defined(DEBUG_STOPWATCH)
    debug(PSTR("start"));
  #endif

  if (isRunning()) return;

  if (isPaused()) accumulator = duration();
  else reset();

  status = STPWTCH_RUNNING;
  startTimestamp = millis();
}

void Stopwatch::reset() {
  #if defined(DEBUG_STOPWATCH)
    debug(PSTR("reset"));
  #endif

  status = STPWTCH_STOPPED;
  startTimestamp = 0;
  stopTimestamp = 0;
  accumulator = 0;
}

bool Stopwatch::isRunning() {
  return (status == STPWTCH_RUNNING) ? true : false;
}

bool Stopwatch::isPaused() {
  return (this->status == STPWTCH_PAUSED) ? true : false;
}

uint16_t Stopwatch::duration() {
  return (((isRunning()) ? millis() : stopTimestamp)
          - startTimestamp) / 1000 + accumulator;
}

void Stopwatch::restart()
{
#if defined(DEBUG_STOPWATCH)
  debug(PSTR("restart"));
#endif

reset();

status = STPWTCH_RUNNING;
startTimestamp = millis();
}


#if defined(DEBUG_STOPWATCH)

  void Stopwatch::debug(const char func[]) {
    if (DEBUGGING(INFO)) {
      SERIAL_ECHOPGM("Stopwatch::");
      serialprintPGM(func);
      SERIAL_ECHOLNPGM("()");
    }
  }

#endif
