/*
    K8400 firmware modification based on original Version provided by Vellemann.

 Copyright (C) 2017 Holger Höhnemann

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */


#include "ultralcd.h"
#include "temperature.h"
#include "watchdog.h"
#include "Configuration.h"
#include "language.h"
#include "printer.h"
#include "MarlinHOH.h"

//===========================================================================
//================================== macros =================================
//===========================================================================

#ifdef K1 // Defined in Configuration.h in the PID settings
#define K2 (1.0-K1)
#endif


//===========================================================================
//=============================private variables============================
//===========================================================================


//#ifdef PIDTEMPBED
////static cannot be external:
//static float temp_iState_bed = { 0 };
//static float temp_dState_bed = { 0 };
//static float pTerm_bed;
//static float iTerm_bed;
//static float dTerm_bed;
////int output;
//static float pid_error_bed;
//static float temp_iState_min_bed;
//
//#else //PIDTEMPBED
////static unsigned long  previous_millis_bed_heater;
//#endif //PIDTEMPBED


#ifdef WATCH_TEMP_PERIOD
int watch_start_temp[EXTRUDERS] = ARRAY_BY_EXTRUDERS(0,0,0);
unsigned long watchmillis[EXTRUDERS] = ARRAY_BY_EXTRUDERS(0,0,0);
#endif //WATCH_TEMP_PERIOD


//===========================================================================
//=============================   Interrupts      ============================
//===========================================================================
ISR(TIMER0_COMPB_vect)
{
	heater.isr_TIMER0_COMPB_vect();
}

ISR (ADC_vect)
{
	heater.isr_ADC_vect();
}

//===========================================================================
//=============================   functions      ============================
//===========================================================================

#if (defined(EXTRUDER_0_AUTO_FAN_PIN) && EXTRUDER_0_AUTO_FAN_PIN > -1) || \
		(defined(EXTRUDER_1_AUTO_FAN_PIN) && EXTRUDER_1_AUTO_FAN_PIN > -1) || \
		(defined(EXTRUDER_2_AUTO_FAN_PIN) && EXTRUDER_2_AUTO_FAN_PIN > -1)

#if defined(FAN_PIN) && FAN_PIN > -1
#if EXTRUDER_0_AUTO_FAN_PIN == FAN_PIN
#error "You cannot set EXTRUDER_0_AUTO_FAN_PIN equal to FAN_PIN"
#endif
#if EXTRUDER_1_AUTO_FAN_PIN == FAN_PIN
#error "You cannot set EXTRUDER_1_AUTO_FAN_PIN equal to FAN_PIN"
#endif
#if EXTRUDER_2_AUTO_FAN_PIN == FAN_PIN
#error "You cannot set EXTRUDER_2_AUTO_FAN_PIN equal to FAN_PIN"
#endif
#endif

void setExtruderAutoFanState(int pin, bool state)
{
	unsigned char newFanSpeed = (state != 0) ? EXTRUDER_AUTO_FAN_SPEED : 0;
	// this idiom allows both digital and PWM fan outputs (see M42 handling).
	pinMode(pin, OUTPUT);
	digitalWrite(pin, newFanSpeed);
	analogWrite(pin, newFanSpeed);
}

void checkExtruderAutoFans()
{
	uint8_t fanState = 0;

	// which fan pins need to be turned on?
#if defined(EXTRUDER_0_AUTO_FAN_PIN) && EXTRUDER_0_AUTO_FAN_PIN > -1
	if (current_temperature[0] > EXTRUDER_AUTO_FAN_TEMPERATURE)
		fanState |= 1;
#endif
#if defined(EXTRUDER_1_AUTO_FAN_PIN) && EXTRUDER_1_AUTO_FAN_PIN > -1
	if (current_temperature[1] > EXTRUDER_AUTO_FAN_TEMPERATURE)
	{
		if (EXTRUDER_1_AUTO_FAN_PIN == EXTRUDER_0_AUTO_FAN_PIN)
			fanState |= 1;
		else
			fanState |= 2;
	}
#endif
#if defined(EXTRUDER_2_AUTO_FAN_PIN) && EXTRUDER_2_AUTO_FAN_PIN > -1
	if (current_temperature[2] > EXTRUDER_AUTO_FAN_TEMPERATURE)
	{
		if (EXTRUDER_2_AUTO_FAN_PIN == EXTRUDER_0_AUTO_FAN_PIN)
			fanState |= 1;
		else if (EXTRUDER_2_AUTO_FAN_PIN == EXTRUDER_1_AUTO_FAN_PIN)
			fanState |= 2;
		else
			fanState |= 4;
	}
#endif

// update extruder auto fan states
#if defined(EXTRUDER_0_AUTO_FAN_PIN) && EXTRUDER_0_AUTO_FAN_PIN > -1
	setExtruderAutoFanState(EXTRUDER_0_AUTO_FAN_PIN, (fanState & 1) != 0);
#endif
#if defined(EXTRUDER_1_AUTO_FAN_PIN) && EXTRUDER_1_AUTO_FAN_PIN > -1
	if (EXTRUDER_1_AUTO_FAN_PIN != EXTRUDER_0_AUTO_FAN_PIN)
		setExtruderAutoFanState(EXTRUDER_1_AUTO_FAN_PIN, (fanState & 2) != 0);
#endif
#if defined(EXTRUDER_2_AUTO_FAN_PIN) && EXTRUDER_2_AUTO_FAN_PIN > -1
	if (EXTRUDER_2_AUTO_FAN_PIN != EXTRUDER_0_AUTO_FAN_PIN
			&& EXTRUDER_2_AUTO_FAN_PIN != EXTRUDER_1_AUTO_FAN_PIN)
		setExtruderAutoFanState(EXTRUDER_2_AUTO_FAN_PIN, (fanState & 4) != 0);
#endif
}

#endif // any extruder auto fan pins set



#define PGM_RD_W(x)   (short)pgm_read_word(&x)

#ifdef HEATER_0_USES_MAX6675
#define MAX6675_HEAT_INTERVAL 250
long max6675_previous_millis = -HEAT_INTERVAL;
int max6675_temp = 2000;

int read_max6675()
{
	if (millis() - max6675_previous_millis < MAX6675_HEAT_INTERVAL)
		return max6675_temp;

	max6675_previous_millis = millis();
	max6675_temp = 0;

#ifdef	PRR
	PRR &= ~(1<<PRSPI);
#elif defined PRR0
	PRR0 &= ~(1<<PRSPI);
#endif

	SPCR = (1<<MSTR) | (1<<SPE) | (1<<SPR0);

	// enable TT_MAX6675
	WRITE(MAX6675_SS, 0);

	// ensure 100ns delay - a bit extra is fine
	asm("nop");//50ns on 20Mhz, 62.5ns on 16Mhz
	asm("nop");//50ns on 20Mhz, 62.5ns on 16Mhz

	// read MSB
	SPDR = 0;
	for (;(SPSR & (1<<SPIF)) == 0;);
	max6675_temp = SPDR;
	max6675_temp <<= 8;

	// read LSB
	SPDR = 0;
	for (;(SPSR & (1<<SPIF)) == 0;);
	max6675_temp |= SPDR;

	// disable TT_MAX6675
	WRITE(MAX6675_SS, 1);

	if (max6675_temp & 4)
	{
		// thermocouple open
		max6675_temp = 2000;
	}
	else
	{
		max6675_temp = max6675_temp >> 3;
	}

	return max6675_temp;
}
#endif


temperatur_Control::temperatur_Control()
{
	 raw_temp_0_value = 0;
#if EXTRUDERS > 1
	 raw_temp_1_value = 0;
#endif
	 raw_temp_bed_value = 0;

	 minttemp_raw[0] = HEATER_0_RAW_LO_TEMP ;
	maxttemp_raw[0] = HEATER_0_RAW_HI_TEMP ;
#if EXTRUDERS > 1
	 minttemp_raw[1] = HEATER_1_RAW_LO_TEMP ;
	maxttemp_raw[1] =  HEATER_1_RAW_HI_TEMP ;
#endif
	bed_minttemp_raw = HEATER_BED_RAW_LO_TEMP;
 bed_maxttemp_raw = HEATER_BED_RAW_HI_TEMP;

	heater_PWM_bed = 0;
	heat_bed = 0;
#ifdef PREVENT_DANGEROUS_EXTRUDE
   allow_cold_extrude = false;
  extrude_min_temp = EXTRUDE_MINTEMP;
#endif
  blink_count = 0;
}



temperatur_Control::~temperatur_Control()
{
}

void temperatur_Control::Init()
{
	// calculate the temperature limits for the different heaters
#ifdef HEATER_0_MINTEMP
//	minttemp[0] = HEATER_0_MINTEMP;
	while(ADC2Temp(minttemp_raw[0], 0) < HEATER_0_MINTEMP) {
#if HEATER_0_RAW_LO_TEMP < HEATER_0_RAW_HI_TEMP
		minttemp_raw[0] += OVERSAMPLENR;
#else
		minttemp_raw[0] -= OVERSAMPLENR;
#endif
	}
#endif //MINTEMP

#ifdef HEATER_0_MAXTEMP
//	maxttemp[0] = HEATER_0_MAXTEMP;
	while(ADC2Temp(maxttemp_raw[0], 0) > HEATER_0_MAXTEMP) {
#if HEATER_0_RAW_LO_TEMP < HEATER_0_RAW_HI_TEMP
		maxttemp_raw[0] -= OVERSAMPLENR;
#else
		maxttemp_raw[0] += OVERSAMPLENR;
#endif
	}
#endif //MAXTEMP

#if (EXTRUDERS > 1) && defined(HEATER_1_MINTEMP)
//	minttemp[1] = HEATER_1_MINTEMP;
	while(ADC2Temp(minttemp_raw[1], 1) < HEATER_1_MINTEMP) {
#if HEATER_1_RAW_LO_TEMP < HEATER_1_RAW_HI_TEMP
		minttemp_raw[1] += OVERSAMPLENR;
#else
		minttemp_raw[1] -= OVERSAMPLENR;
#endif
	}
#endif // MINTEMP 1

#if (EXTRUDERS > 1) && defined(HEATER_1_MAXTEMP)
//	maxttemp[1] = HEATER_1_MAXTEMP;
	while(ADC2Temp(maxttemp_raw[1], 1) > HEATER_1_MAXTEMP) {
#if HEATER_1_RAW_LO_TEMP < HEATER_1_RAW_HI_TEMP
		maxttemp_raw[1] -= OVERSAMPLENR;
#else
		maxttemp_raw[1] += OVERSAMPLENR;
#endif
	}
#endif //MAXTEMP

#if (TEMP_SENSOR_BED != 0) && defined(BED_MINTEMP)
  while(ADC2Temp(bed_minttemp_raw,HEAT_BED) < BED_MINTEMP) {
#if HEATER_BED_RAW_LO_TEMP < HEATER_BED_RAW_HI_TEMP
    bed_minttemp_raw += OVERSAMPLENR;
#else
    bed_minttemp_raw -= OVERSAMPLENR;
#endif
  }
#endif //BED_MINTEMP

#ifdef BED_MAXTEMP
	while(ADC2Temp(bed_maxttemp_raw,HEAT_BED) > BED_MAXTEMP) {
#if HEATER_BED_RAW_LO_TEMP < HEATER_BED_RAW_HI_TEMP
		bed_maxttemp_raw -= OVERSAMPLENR;
#else
		bed_maxttemp_raw += OVERSAMPLENR;
#endif
	}
#endif //BED_MAXTEMP

//#ifdef PIDTEMPBED
//		temp_iState_min_bed = 0.0;
//		temp_iState_max_bed = PID_INTEGRAL_DRIVE_MAX / bedKi;
//#endif //PIDTEMPBED
////	}

		// set heater pins
#if defined(HEATER_0_PIN) && (HEATER_0_PIN > -1)
	SET_OUTPUT(HEATER_0_PIN);
#endif

#if defined(HEATER_1_PIN) && (HEATER_1_PIN > -1)
	SET_OUTPUT(HEATER_1_PIN);
#endif

#if defined(HEATER_BED_PIN) && (HEATER_BED_PIN > -1)
	SET_OUTPUT(HEATER_BED_PIN);
#endif

#if defined(FAN_PIN) && (FAN_PIN > -1)
	SET_OUTPUT(FAN_PIN);
	sbi (TCCR4A,WGM40);
#endif

	// Set analog inputs and enable interrupt
	ADCSRA = 1<<ADEN | 1<<ADSC | 1<<ADIE | 0x07;
	DIDR0 = 0;
#ifdef DIDR2
	DIDR2 = 0;
#endif
#if defined(TEMP_0_PIN) && (TEMP_0_PIN > -1)
#if TEMP_0_PIN < 8
	DIDR0 |= 1 << TEMP_0_PIN;
#else
	DIDR2 |= 1<<(TEMP_0_PIN - 8);
#endif
#endif
#if defined(TEMP_1_PIN) && (TEMP_1_PIN > -1)
#if TEMP_1_PIN < 8
	DIDR0 |= 1<<TEMP_1_PIN;
#else
	DIDR2 |= 1<<(TEMP_1_PIN - 8);
#endif
#endif
#if defined(TEMP_2_PIN) && (TEMP_2_PIN > -1)
#if TEMP_2_PIN < 8
	DIDR0 |= 1 << TEMP_2_PIN;
#else
	DIDR2 |= 1<<(TEMP_2_PIN - 8);
#endif
#endif
#if defined(TEMP_BED_PIN) && (TEMP_BED_PIN > -1)
#if TEMP_BED_PIN < 8
	DIDR0 |= 1<<TEMP_BED_PIN;
#else
	DIDR2 |= 1<<(TEMP_BED_PIN - 8);
#endif
#endif

	// initialisae and start ADC-Conversion
	for (int i=0; i<16; i++)
		ADC_Temp0[i] = ADC_Temp1[i] = ADC_TempBed[i] = 0;

	ADC_Lauf = 0;
	sei();
	sbi (ADCSRA ,ADSC ); // Start conversion

	// Use timer0 for temperature measurement
	// Interleave temperature interrupt with millies interrupt
	OCR0B = 128;
	TIMSK0 |= (1<<OCIE0B);

	// Wait for temperature measurement to settle
	delay(250);


	UpdatePID();
}

inline void temperatur_Control::isr_TIMER0_COMPB_vect()
{
	lcd_buttons_update();
//	LCD.Check_Buttons();
	temp_count++;

	if(temp_count >= OVERSAMPLENR * 8) // 8 * 16 * 1/(16000000/64/256)  = 131ms.
	{
		temp_meas_ready = true;
		temp_count = 0;

	}
	if (++blink_count > 976)
	{
		blink_count = 0;
		my_printer.blinker = my_printer.blinker ^ true;
	}
}

void temperatur_Control::Disable()
{
#if defined(TEMP_BED_PIN) && TEMP_BED_PIN > -1
	for(int i=HEAT_BED;i<EXTRUDERS;i++)
#else
		for(int i=0;i<EXTRUDERS;i++)
#endif
	{
		setTargetHeater(0,i);
		Set_Heater(i,0);
	}
}

void temperatur_Control::Manage()
{
	float pid_output;

	// handle error-Messages first

	switch (heater_error)
	{
	case No_Error:
		break;
	case MinTempBed:
	case MaxTempBed:
		break;
	case MinTemp0:
		Min_Temp_Error(0);
		break;
	case MaxTemp0:
		if (error_count++ > 20)
			if (fanSpeed)
			{
				Set_FanSpeed(fanSpeed-1);
				error_count = 0;
			}
		Max_Temp_Error(0);
		break;
	case MinTemp1:
		Min_Temp_Error(1);
		break;
	case MaxTemp1:
		if (error_count++ > 20)
			if (fanSpeed)
			{
				Set_FanSpeed(fanSpeed-1);
				error_count = 0;
			}
		Max_Temp_Error(1);
		break;
	default:
		my_printer.error_ln(PSTR("undefined error code"));
	}
	heater_error = 0;

	if(temp_meas_ready != true)   //better readability
		return;
	UpdateTemperaturesFromRawValues(); // resets "temp_meas_ready"

#ifdef PIDTEMPBED
	for(int e = HEAT_BED; e < EXTRUDERS; e++)
#else
	for(int e = -1; e < EXTRUDERS; e++)
#endif
	{

#ifndef PID_OPENLOOP

		pid_error[e+1] = degTargetHotend(e) - degHotend(e);
//
//		kappa = config.kd(e) / config.kp(e);
//		kappa = kappa/(kappa+PID_N);

		//		dTerm[e] = K2 * PID_PARAM(Kd, e) * (current_temperature[e] - temp_dState[e]) + K1 * dTerm[e];
		dTerm[e+1] =  kappa[e+1] *( config.kp(e) * PID_N *(degHotend(e) - temp_dState[e+1]) - dTerm[e+1]);
		temp_dState[e+1] = degHotend(e);
		if(pid_error[e+1] > PID_FUNCTIONAL_RANGE)
		{
			pid_output = PID_MAX;
			pid_reset[e+1] = true;
		}
		else if(pid_error[e+1] < -PID_FUNCTIONAL_RANGE || degTargetHotend(e) == 0)
		{
			pid_output = 0;
			pid_reset[e+1] = true;
		}
		else
		{
			if(pid_reset[e+1] == true)
			{
				temp_iState[e+1] = 0.0;
				pid_reset[e+1] = false;
			}

			pTerm[e+1] = config.kp(e) * pid_error[e+1];
			temp_iState[e+1] += config.ki(e) *pid_error[e+1];
			temp_iState[e+1] = constrain(temp_iState[e+1], temp_iState_min[e+1], temp_iState_max[e+1]);
			iTerm[e+1] =  temp_iState[e+1];

			pid_output = pTerm[e+1] + iTerm[e+1] - dTerm[e+1];

#ifdef PID_ADD_EXTRUSION_RATE
			cTerm[e+1] = 0;
			// add fan dependant parameter
			cTerm[e+1] = config.kf(e)* fanSpeed ;
			if (cTerm[e+1] > 0)
				cTerm[e+1]+= degTargetHotend(e)/3;
			if (e == active_extruder)
			{
				move_block *block = plan.Get_Current_Block();

				if (block)
				{
					if (block->steps[E_AXIS] && (block->direction_bits &(1<<E_AXIS))==0)
						// nominal speed is extruder speed if extruder is active
					cTerm[e+1] +=  block->nominal_speed * config.kc(e) ;
				}
			}
			pid_output += cTerm[e+1];
#endif //PID_ADD_EXTRUSION_RATE
		}
#else
		pid_output = constrain(target_temperature[e], 0, PID_MAX);
#endif //PID_OPENLOOP

		{
//			if (! pid_reset[e] && pid_output > PID_MAX)
//				LCD_ALERTMESSAGEPGM("Err: HEATER OVERFLOW");
			Set_Heater(e, pid_output);
		}
	}

} // End extruder for loop

// todo: Implementation of BED PID-Control

void temperatur_Control::UpdatePID()
{
#ifdef PIDTEMPBED
	for(int e = HEAT_BED; e < EXTRUDERS; e++)
#else
	for(int e = 0; e < EXTRUDERS; e++)
#endif
	{
		temp_iState_max[e+1] = PID_INTEGRAL_DRIVE_MAX / config.ki(e);
		temp_iState_min[e+1] = 0.0;
		kappa[e+1]= config.kd(e) / config.kp(e);
		kappa[e+1] = kappa[e+1]/(kappa[e+1]+PID_N);	}

}


inline void temperatur_Control::isr_ADC_vect()
{
	switch (ADC_state)
	{
	case -1:
		// start bed temperature measurement
#if defined(TEMP_BED_PIN) && (TEMP_BED_PIN > -1)
#if TEMP_BED_PIN > 7
		ADCSRB = 1<<MUX5;
#else
		ADCSRB = 0;
#endif
		ADMUX = ((1 << REFS0) | (TEMP_BED_PIN & 0x07));
		ADC_state++;

		sbi (ADCSRA ,ADSC );
		ADCSRA |= 1<<ADSC; // Start conversion
#endif
		break;
	case 0:
		////	read temp_bed and prepare extruder 1 temp measurementr
#if defined(TEMP_0_PIN) && (TEMP_0_PIN > -1)
#if TEMP_0_PIN > 7
		ADCSRB = 1<<MUX5;
#else
		ADCSRB = 0;
#endif
		ADMUX = ((1 << REFS0) | (TEMP_0_PIN & 0x07));
		ADC_TempBed[ADC_Lauf]= ADC;
//		ADC_TempBed[ADC_Lauf]= 400;  // dummy value to avoid cross coupling


//		if (ADC > minttemp_raw[0])
//		{
//			Set_Heater(0,0);
//			heater_error = MinTemp0;
//		}
//		if (ADC < maxttemp_raw[0])
//		{
//			Set_Heater(0,0);
//			heater_error = MaxTemp0;
//		}
		ADC_state++;
//		ADCSRA |= 1<<ADSC; // Start conversion
		sbi (ADCSRA ,ADSC ); // Start conversion
#endif
		break;
	case 1:
		////	read extruder 1 dummy measurement and prepare extruder 1 measurementr
#if defined(TEMP_0_PIN) && (TEMP_0_PIN > -1)
#if TEMP_0_PIN > 7
		ADCSRB = 1<<MUX5;
#else
		ADCSRB = 0;
#endif
		ADMUX = ((1 << REFS0) | (TEMP_0_PIN & 0x07));

		ADC_state++;
		sbi (ADCSRA ,ADSC ); // Start conversion
#endif
		break;
	case 2:
		////	read extruder 1  measurement and prepare extruder 2 dummy measurementr


#if defined(TEMP_1_PIN) && (TEMP_1_PIN > -1)
#if TEMP_1_PIN > 7
		ADCSRB = 1<<MUX5;
#else
		ADCSRB = 0;
#endif
		ADMUX = ((1 << REFS0) | (TEMP_1_PIN & 0x07));
		ADC_Temp0[ADC_Lauf]= ADC;
				if (ADC<<4 > minttemp_raw[0])
				{
					Set_Heater(0,0);
					heater_error = MinTemp0;
				}
				if (ADC<<4 < maxttemp_raw[0])
				{
					Set_Heater(0,0);
					heater_error = MaxTemp0;
				}

//		if (ADC > minttemp_raw[0])
//		{
//			Set_Heater(HEAT_BED,0);
//			heater_error = MinTemp0;
//		}
//		if (ADC < maxttemp_raw[0])
//		{
//			Set_Heater(HEAT_BED,0);
//			heater_error = MaxTemp0;
//		}
		ADC_state++;
		ADCSRA |= 1<<ADSC; // Start conversion
		sbi (ADCSRA ,ADSC ); // Start conversion
#endif

		break;
	case 3:
		////	read extruder 2  dummy measurement and prepare extruder 2  measurementr

#if defined(TEMP_1_PIN) && (TEMP_1_PIN > -1)
#if TEMP_1_PIN > 7
		ADCSRB = 1<<MUX5;
#else
		ADCSRB = 0;
#endif
		ADMUX = ((1 << REFS0) | (TEMP_1_PIN & 0x07));

		ADC_state++;
		ADCSRA |= 1<<ADSC; // Start conversion
		sbi (ADCSRA ,ADSC ); // Start conversion
#endif

		break;
	case 4:
		////	read extruder 2   measurement and prepare bed dummy measurementr

#if defined(TEMP_1_PIN) && (TEMP_1_PIN > -1)
#if TEMP_1_PIN > 7
		ADCSRB = 1<<MUX5;
#else
		ADCSRB = 0;
#endif
		ADMUX = ((1 << REFS0) | (TEMP_BED_PIN & 0x07));
		ADC_Temp1[ADC_Lauf]= ADC;
		if (ADC<<4 > minttemp_raw[0])
		{
			Set_Heater(1,0);
			heater_error = MinTemp1;
		}
		if (ADC<<4 < maxttemp_raw[0])
		{
			Set_Heater(1,0);
			heater_error = MaxTemp1;
		}
		ADC_state++;
		ADCSRA |= 1<<ADSC; // Start conversion
		sbi (ADCSRA ,ADSC ); // Start conversion
#endif

		break;
	default:
		// start dummy bed temperature measurement
#if defined(TEMP_BED_PIN) && (TEMP_BED_PIN > -1)
#if TEMP_BED_PIN > 7
		ADCSRB = 1<<MUX5;
#else
		ADCSRB = 0;
#endif
		ADMUX = ((1 << REFS0) | (TEMP_BED_PIN & 0x07));

		ADC_state = 0;
		ADC_Lauf = ++ADC_Lauf % 16;
		ADCSRA |= 1<<ADSC; // Start conversion
		sbi (ADCSRA ,ADSC );
#endif
		break;
	}
}

float temperatur_Control::ADC2Temp(int e)
{

	int raw = 0;

	if(e >= EXTRUDERS)
	{
		my_printer.f_message(T(MSG_Invalid_Extruder_Num),e);
		kill();
	}
	//  read raw values
	CRITICAL_SECTION_START;
	switch (e)
	{
	case HEAT_BED:
		for (int i=0; i<16; i++)
			raw += ADC_TempBed[i];
		break;
	case 0:
		for (int i=0; i<16; i++)
			raw += ADC_Temp0[i];
		break;
	case 1:
		for (int i=0; i<16; i++)
			raw += ADC_Temp1[i];
		break;
	}
	CRITICAL_SECTION_END;
	return ADC2Temp(raw, e);

}



void temperatur_Control::autotempShutdown()
{
	//	 #ifdef AUTOTEMP
//	if(autotemp_enabled)
//	{
//		autotemp_enabled=false;
//		if(degTargetHotend(active_extruder)>autotemp_min)
//			setTargetHeater(0,active_extruder);
//	}
//	//	 #endif
}

inline void temperatur_Control::Print_Heater_States()
{
#if defined(TEMP_0_PIN) && TEMP_0_PIN > -1
	my_printer.f_message(PSTR("ok T:%.1f /%.0f"),degHotend(active_extruder),degTargetHotend(active_extruder));

#endif
#if  defined(TEMP_BED_PIN) && TEMP_BED_PIN > -1
	my_printer.f_message(PSTR(" B:%.1f /%.1f"),degHotend(HEAT_BED),degTargetHotend(HEAT_BED));
#endif
#if EXTRUDERS > 1
	for (int8_t e = 0; e < EXTRUDERS; ++e)
	{
		my_printer.f_message(PSTR(" T%i:%.1f /%.0f"),e,degHotend(e),degTargetHotend(e));
	}
#endif

	my_printer.f_message(PSTR(" B@:%i @:%i"),Get_Power(-1),Get_Power(active_extruder));
#if EXTRUDERS > 1
	for (int8_t e = 0; e < EXTRUDERS; ++e)
	{
		my_printer.f_message(PSTR(" @%i:%i"),e,Get_Power(e));
	}
#endif
	my_printer.message(PSTR("\n"));
}

inline int temperatur_Control::Set_FanSpeed(int speed)
{
	fanSpeed = constrain(speed,0,255);
	Write_Analogue(FAN_PIN, fanSpeed);
    return fanSpeed;
}

void temperatur_Control::UpdateTemperaturesFromRawValues()
{
	/* Called to get the raw values into the the actual temperatures. The raw values are created in interrupt context,
	    and this function is called from normal context as it is too slow to run in interrupts and will block the stepper routine otherwise */

		for(uint8_t e=0;e<EXTRUDERS;e++)
		{
			current_temperature[e] = ADC2Temp(e);
		}
		current_temperature_bed = ADC2Temp(HEAT_BED);

		//Reset the watchdog after we know we have a temperature measurement.
		watchdog_reset();

		CRITICAL_SECTION_START;
		temp_meas_ready = false;
		CRITICAL_SECTION_END;

}

int temperatur_Control::Get_Power(int extruder)
{
	if (extruder<0)
		return heater_PWM_bed;
	if (extruder < EXTRUDERS)
		return heater_PWM[extruder];
	return 0;
}


unsigned char temperatur_Control::Set_Heater(int e, int value)
{
	switch (e)
	{
	case -1:

		heater_PWM_bed = constrain(value,0,255);
		Write_Analogue(HEATER_BED_PIN, heater_PWM_bed);
		heat_bed += heater_PWM_bed;
		return heater_PWM_bed;
		break;
	case 0:

		heater_PWM[e] = constrain(value,0,255);
		Write_Analogue(HEATER_0_PIN, heater_PWM[e]);
		heat[e] +=  heater_PWM[e];
		return heater_PWM[e];
		break;
	case 1:

		heater_PWM[e] = constrain(value,0,255);
		Write_Analogue(HEATER_1_PIN, heater_PWM[e]);
		heat[e] +=  heater_PWM[e];
		return heater_PWM[e];
		break;
	}
	return 0;
}

void temperatur_Control::PID_autotune(float temp, int extruder, int ncycles,
		bool update)
{

	float input = 0.0;
	int cycles=0;
	bool heating = true;
	float SKu = 0;
	float STu = 0;

	unsigned long temp_millis = millis();
	unsigned long t1=temp_millis;
	unsigned long t2=temp_millis;
	long t_high = 0;
	long t_low = 0;

	int bias, d;
	float Ku = 0;
	float Tu = 0;
	float nKp, nKi, nKd;
	float max = 0, min = 10000;

	bool	M105_request = false;

	if ((extruder > EXTRUDERS)
#if (TEMP_BED_PIN <= -1)
			||(extruder < 0)
#endif
	){
		my_printer.error_ln(T(MSG_PID_Autotune_Invalid_Extrude));
		return;
	}

	my_printer.echo_ln(T(MSG_PID_Autotune_Start));

	Disable(); // switch off all heaters.

	// cooldown to max 100°C
	timer.start();
	while (degHotend(extruder) > 100 && extruder >= 0)
	{
		//          switch on the fan
		Set_FanSpeed( 255);
		Manage();
		if (timer.duration() > 1)
		{
			Print_Heater_States();
			timer.restart();
		}
		lcd_update();
	}
	timer.reset();
	// 			temperature sufficient low
	//          switch off the fan
	Set_FanSpeed(0);

	// Start with step response

	if (extruder>=0)
	{
		float slope [5] = {0,0,0,0,0};
		float mean_slope = 0;
		float error;
		float temp_0;
		float delta_t = 0;
		unsigned long last;
		float temp_T = 0;
		for (int i=0; i< 5; i++)
		{
			slope[i] = 0;
		}
		my_printer.message(T(MSG_Hotend_Stabilise));

		Set_Heater(extruder,PID_MAX/50);

		temp_0 = temp_T = Wait_T_Stable(extruder,false);
		my_printer.f_message(T(MSG_Hotend_Stable_Temperature),temp_T);

		// heat step start
		t1 = temp_millis = last = millis();

		Set_Heater(extruder,PID_MAX);
		max = -1;
		while (heating)
		{
				t2 = millis();
			if((temp_meas_ready == true) && ((t2 - temp_millis)>1800))
			{ // temp sample ready
				UpdateTemperaturesFromRawValues();
				temp_meas_ready = false;
				input = (extruder<0)?current_temperature_bed:degHotend(extruder);
				t2 = millis();

				delta_t = (t2 - temp_millis)/1000.0;
				slope[0] = slope[1];
				slope[1] = slope[2];
				slope[2] = slope[3];
				slope[3] = slope[4];
				slope[4]= (input - temp_T)/delta_t;

				mean_slope = (slope[0] + slope[1] + slope[2] + slope[3]+slope[4] )/5.0;
				error = (slope[1]-mean_slope + slope[2]-mean_slope + slope[3]-mean_slope +slope[4]-mean_slope + slope[0]-mean_slope)+0.0001;

				if (mean_slope < (max - abs(error)) && (mean_slope > 0.1))
				{
					// end of test, use last values to calculate parameters
					Set_Heater(extruder,0);;

					// calculate  the characteristic values
					Ku = max ;
					Tu = (temp_millis - t1)/1000.0 - (temp_T - temp_0)/max;
					heating = false;
				}
				else
				{
					max = mean_slope;
					temp_T = input ;
				}
				Print_Heater_States();
				temp_millis = t2;
			}

			lcd_update();

		}
		my_printer.f_message(PSTR("Slope: %f  Delay: %f\n"),Ku,Tu);

		nKp = 1.2*255/(Ku*Tu);
		nKi = 0.5*nKp/Tu;
		nKd = 0.3*nKp*Tu;
		my_printer.message(PSTR(" Classic PID \n"));
		my_printer.f_message(PSTR(" Kp: %f\n Ki: %f\n Kd: %f\n"),nKp, nKi, nKd);



	}

	max = 0;
	min = 1000;
	temp_millis = millis();
	t1=temp_millis;
	t2=temp_millis;

	if (extruder<0)
	{
		Set_Heater(HEAT_BED,MAX_BED_POWER);

		bias = d = (MAX_BED_POWER);
	}
	else
	{
		Set_Heater(extruder,PID_MAX);
		bias = d = PID_MAX;
	}
	heating = true;

	// PID Tuning loop
	my_printer.message(PSTR("PID relay tuning\n"));

	for(;;)
	{
		if(temp_meas_ready == true)
		{ // temp sample ready
			UpdateTemperaturesFromRawValues();
			temp_meas_ready = false;
			input = (extruder<0)?current_temperature_bed:degHotend(extruder);

			max=max(max,input);
			min=min(min,input);
			if(heating == true && input > temp)
			{
				if(millis() - t2 > 5000)
				{
					heating=false;
					if (extruder<0)
						Set_Heater(HEAT_BED,(bias - d) );
					else
						Set_Heater(extruder,(bias - d));
					t1=millis();
					t_high=t1 - t2;
					max=temp;
				}
			}
			if(heating == false && input < temp) {
				if(millis() - t1 > 5000) {
					heating=true;
					t2=millis();
					t_low=t2 - t1;
					if(cycles > 0)
					{
						bias += (d*(t_high - t_low))/(t_low + t_high);
						bias = constrain(bias, 20 ,(extruder<0?(MAX_BED_POWER):(PID_MAX))-90);
						if(bias > (extruder<0?(MAX_BED_POWER):(PID_MAX))/2)
							d = (extruder<0?(MAX_BED_POWER):(PID_MAX)) - 1 - bias;
						else
							d = bias;

						my_printer.f_message(PSTR("bias: %i d: %i min: %.1f max: %.1f\n"),bias,d,min,max);

						if(cycles > 2) {
							Ku = (4.0*d)/(3.14159*(max-min));
							SKu += Ku;
							Tu = ((float)(t_low + t_high)/1000.0);
							STu += Tu;
							my_printer.f_message(PSTR("Ku: %f  Tu: %f\n"),Ku,Tu);

							nKp = 0.5*Ku;
							nKi = nKp/Tu;
							nKd = 0.4*nKp*Tu;
							my_printer.message(PSTR(" Classic PID \n"));
							my_printer.f_message(PSTR(" Kp: %f\n Ki: %f\n Kd: %f\n"),nKp, nKi, nKd);

						}
					}
					if (extruder<0)
						Set_Heater(HEAT_BED,(bias + d) );
					else
						Set_Heater(extruder,(bias + d));
					cycles++;
					min=temp;
				}
			}
		}
		if(input > (temp + 20)) {
			my_printer.error_ln(T(MSG_PID_Autotune_Failed));
			return;
		}
		if(millis() - temp_millis > 2000)
		{
			Print_Heater_States();
			if (M105_request)
			{
				  my_printer.message(T(MSG_OK));
//				SERIAL_PROTOCOLPGM(MSG_OK);
				M105_request = false;
			}
			temp_millis = millis();
		}
		if(((millis() - t1) + (millis() - t2)) > (20lu*60lu*1000lu*2lu))
		{
			my_printer.message(T(MSG_PID_Autotune_Timeout));
			return;
		}
		if(cycles > ncycles)
		{
			Ku = SKu / (ncycles - 2.0);
			Tu = STu / (ncycles - 2.0);
			my_printer.f_message(PSTR("Ku: %f  Tu: %f\n"),Ku,Tu);;
			nKp = 0.5*Ku;
			nKi = nKp/Tu;
			nKd = 0.3*nKp*Tu;
			my_printer.message(PSTR(" Classic PID \n"));
			my_printer.f_message(PSTR(" Kp: %f\n Ki: %f\n Kd: %f\n"),nKp, nKi, nKd);
			if (update)
			{
				config.Set_Kp(nKp,extruder);
				config.Set_Ki(nKi,extruder);
				config.Set_Kd(nKd,extruder);
				my_printer.message(T(MSG_PID_Autotune_Finished_Updated));
				UpdatePID();
			}
			else
			{
				my_printer.message(T(MSG_PID_Autotune_Finished));
			}
			return;
		}
		lcd_update();
	}
}

float temperatur_Control::ADC2Temp(int value, int e)
{
	switch (e)
	{
	case -1:
	{
#if TEMP_SENSOR_BED > 0

		if(heater_ttbl_map_bed != NULL)
		{
			float celsius = 0;
			uint8_t i;
			short (*tt)[][2] = (short (*)[][2])(heater_ttbl_map_bed);

			for (i=1; i<heater_ttbllen_map_bed; i++)
			{
				if (PGM_RD_W((*tt)[i][0]) > value)
				{
					celsius = PGM_RD_W((*tt)[i-1][1]) +
							(value - PGM_RD_W((*tt)[i-1][0])) *
							(float)(PGM_RD_W((*tt)[i][1]) - PGM_RD_W((*tt)[i-1][1])) /
							(float)(PGM_RD_W((*tt)[i][0]) - PGM_RD_W((*tt)[i-1][0]));
					break;
				}
			}
			// Overflow: Set to last value in the table
			if (i == heater_ttbllen_map_bed)
				celsius = PGM_RD_W((*tt)[i-1][1]);
			return celsius;
		}
		return ((value * ((5.0 * 100.0) / 1024.0) / OVERSAMPLENR) * TEMP_SENSOR_AD595_GAIN) + TEMP_SENSOR_AD595_OFFSET;
#else
		{
			return 0.0;
		}
#endif
		break;
	}
	case 0:
#if EXTRUDERS > 1
	case 1:
#endif
	{
		if(heater_ttbl_map[e] != NULL)
		{
			float celsius = 0;
			uint8_t i;
			short (*tt)[][2] = (short (*)[][2])(heater_ttbl_map[e]);

			for (i=1; i<heater_ttbllen_map[e]; i++)
			{
				if (PGM_RD_W((*tt)[i][0]) > value)
				{
					celsius = PGM_RD_W((*tt)[i-1][1]) +
							(value - PGM_RD_W((*tt)[i-1][0])) *
							(float)(PGM_RD_W((*tt)[i][1]) - PGM_RD_W((*tt)[i-1][1])) /
							(float)(PGM_RD_W((*tt)[i][0]) - PGM_RD_W((*tt)[i-1][0]));
					break;
				}
			}
			// Overflow: Set to last value in the table
			if (i == heater_ttbllen_map[e])
				celsius = PGM_RD_W((*tt)[i-1][1]);
			return celsius;
		}
		return ((value * ((5.0 * 100.0) / 1024.0) / OVERSAMPLENR) * TEMP_SENSOR_AD595_GAIN) + TEMP_SENSOR_AD595_OFFSET;
	}
	break;
	default:
	{
		my_printer.f_message(T(MSG_Invalid_Extruder_Num),e);
		kill();
		return 0.0;
	}
	}
}

float temperatur_Control::Wait_T_Stable(int selected_extruder,bool heater_control)
{
	float temp, s_temp;
	float input;
	unsigned long last = millis();
	unsigned long first = last  ;
	unsigned long temp_millis = last;
	unsigned long initial = last;
	long	n = 1;
	bool target = heater_control;

	s_temp = temp = heater_control? target_temperature[selected_extruder] : current_temperature[selected_extruder];

	while (true) // ten seconds stable
	{
		if (temp_meas_ready == true)
		{
			if (heater_control)
			{
				Manage();
			}
			else
			{
				UpdateTemperaturesFromRawValues();
				temp_meas_ready = false;
			}
			input = current_temperature[selected_extruder];
			last =	millis();

			if (abs(temp - input)< 2)
			{
				s_temp += input;
				n++;
			}
			else
			{
				first = last;
				s_temp = temp = target? target_temperature[selected_extruder] : current_temperature[selected_extruder];
				n = 1;

			}
			if(last - temp_millis > 2000)
			{
				Print_Heater_States();
				temp_millis = last;
			}
			if ((last - first) > 10000)
				return (s_temp/(n*1.0));

			if ((last - initial) > 50000)  // recovery from constant offset
				target = false;
		}

		lcd_update();
	}
	return -1.0;  // should never occur
}

inline unsigned long temperatur_Control::Heat(int selected_extruder, unsigned long duration)
{
	unsigned long temp_millis = millis();

	timer.start();
	switch (selected_extruder)
	{
	case 0:
		heat[0] = 0;
		break;
	case 1:
		heat[1] = 0;
		break;
	default:
		heat_bed = 0;
	}
	while (timer.duration() < duration/1000)
	{
		if (temp_meas_ready == true)
		{
			heater.Manage();
			if(millis() - temp_millis > 2000)
			{
				Print_Heater_States();
				temp_millis = millis();
			}
		}
	lcd_update();
	}
	timer.reset();
	switch (selected_extruder)
	{
	case 0:
		return heat[0] ;
		break;
	case 1:
		return heat[1];
		break;
	default:
		return heat_bed;
	}
	return 0;

}

void temperatur_Control::Max_Temp_Error(uint8_t e)
{
	heater.Disable();
	if(IsStopped() == false)
	{
		my_printer.error();
		my_printer.f_message(T(MSG_Maxtemp_Error_Extr),e);
		lcd_setalertstatuspgm(T(MSG_Error_Maxtemp));
	}
#ifndef BOGUS_TEMPERATURE_FAILSAFE_OVERRIDE
	Stop();
#endif
}

void temperatur_Control::Min_Temp_Error(uint8_t e)
{
	if(IsStopped() == false) {
		my_printer.error();
		my_printer.f_message(T(MSG_Mintemp_Error_Extr),e);
		lcd_setalertstatuspgm (T(MSG_Error_Mintemp));
	}
#ifndef BOGUS_TEMPERATURE_FAILSAFE_OVERRIDE
	Stop();
#endif
}

bool temperatur_Control::Tune_Extruder_Fan(int selected_extruder, bool update)
{
	//
	// for power correction in the PID routine the following rule is implemented:
	//
	//		p(fan) = Temp(setpoint)/3 + Kf * fanspeed
	//
	//		within this routine the faktor Kf is evaluated.
	//
	//		The procedure is returning "true" when the tuning was properly finished.
	//		If heater overrun (heater settings > PID_MAX) was found, the routine stops, returning "false"
	//		and the default values are set
	//
	int cycles = 0;
	unsigned long heat, ref_heat;
	float delta;
	float kf, kf_50, kf_75;
	float temp = 0;

	//          switch off fan
	Set_FanSpeed(0);
	timer.reset();
	my_printer.message(T(MSG_Fan_Autotune_Started));
	timer.start();
	while (timer.duration() < 30) // settling time for PID
	{
		Manage();
		lcd_update();
	}
	timer.reset();
	temp = Wait_T_Stable(selected_extruder,true);

	cycles = 0;
	while ((abs(target_temperature[selected_extruder]-temp)> 2) && (cycles < 3))
	{
		my_printer.message(T(MSG_Repeat_Temperature_Stabilisation));
		temp = Wait_T_Stable(selected_extruder,true);
		cycles++;
	}
	my_printer.f_message(T(MSG_Hotend_Stable_Temperature),temp);

	ref_heat = Heat ( selected_extruder, 5000);
	my_printer.f_message(PSTR("ref heat: %i\n"),ref_heat);

	// start with 50% fan
	Set_FanSpeed(128);
	timer.start();
	while (timer.duration() < 20) // settling time for PID
	{
		Manage();
		lcd_update();
	}
	timer.reset();
	temp = Wait_T_Stable (selected_extruder,true);
	cycles = 0;
	while ((abs(target_temperature[selected_extruder]-temp)> 2) && (cycles < 3))
	{
		my_printer.message(T(MSG_Repeat_Temperature_Stabilisation));
		temp = Wait_T_Stable(selected_extruder,true);
		cycles++;
	}
	if (cycles < 3)
	{
		my_printer.f_message(T(MSG_Hotend_Stable_Temperature),temp);

		heat = Heat ( selected_extruder, 5000);
		my_printer.f_message(PSTR(" heat: %i\n"),heat);
		delta = ((float)(heat -ref_heat)*PID_dT)/5.0 ;
		kf_50 = (delta- target_temperature[selected_extruder]/4)/(float)fanSpeed;
		my_printer.f_message(PSTR(" f(50%%): %f \n"),kf_50);
	}
	else
	{
		// reduction of tuning, decrease fan speed
		Set_FanSpeed(fanSpeed-16);
		temp = Wait_T_Stable(selected_extruder,true);
		while (abs(temp  -target_temperature[selected_extruder]) > 3)
		{
			Set_FanSpeed(fanSpeed-16);
			temp = Wait_T_Stable(selected_extruder,true);
		}
		my_printer.f_message(T(MSG_Hotend_Stable_Temperature),temp);

		heat = Heat ( selected_extruder, 5000);
		my_printer.f_message(PSTR(" heat: %i\n"),heat);
		delta = ((float)(heat -ref_heat)*PID_dT)/5.0 ;
		kf = (delta- target_temperature[selected_extruder]/4)/(float)fanSpeed;
		my_printer.f_message(PSTR(" f(%i%%): %f \n"),fanSpeed*100/255, kf);

		my_printer.f_message(PSTR("Kf = %f\n"),kf);

		if (update)
			config.Set_Kf(kf,selected_extruder);
		my_printer.f_message(T(MSG_Temperature_Stabilisation_Abort_At),temp);
		//          switch off fan
		Set_FanSpeed(0);
		UpdatePID();
		return false;
	}

	//repeat with 75% fan speed
	Set_FanSpeed(192);
	timer.start();
	while (timer.duration() < 20) // settling time for PID
	{
		Manage();
		lcd_update();
	}
	timer.reset();
	my_printer.f_message(T(MSG_PID_Reset));
	temp = Wait_T_Stable (selected_extruder,true);
	cycles = 0;
	while ((abs(target_temperature[selected_extruder]-temp)> 2) && (cycles < 3))
	{
		my_printer.message(T(MSG_Repeat_Temperature_Stabilisation));
		temp = Wait_T_Stable(selected_extruder,true);
		cycles++;
	}
	if (cycles < 3)
	{

		my_printer.f_message(T(MSG_Hotend_Stable_Temperature),temp);

		heat = Heat ( selected_extruder, 5000);
		my_printer.f_message(PSTR(" heat: %i\n"),heat);
		delta = ((float)(heat -ref_heat)*PID_dT)/5.0;
		kf_75 = (delta - target_temperature[selected_extruder]/4)/(float)fanSpeed;
		my_printer.f_message(PSTR(" f(75%%): %f \n"),kf_75);
	}
	else
	{
		my_printer.f_message(T(MSG_Temperature_Stabilisation_Abort_At),temp);
		//          switch off fan
		Set_FanSpeed(0);
		return false;
	}
	//repeat with 100% fan speed
	Set_FanSpeed(255);
	timer.start();
	while (timer.duration() < 30) // settling time for PID
	{
		Manage();
		lcd_update();
	}
	timer.reset();
	temp = Wait_T_Stable (selected_extruder,true);
	cycles = 0;
	while ((abs(target_temperature[selected_extruder]-temp)> 2) && (cycles < 3))
	{
		my_printer.message(T(MSG_Repeat_Temperature_Stabilisation));
		temp = Wait_T_Stable(selected_extruder,true);
		cycles++;
	}
	if (cycles < 3)
	{

		my_printer.f_message(T(MSG_Hotend_Stable_Temperature),temp);

		heat = Heat ( selected_extruder, 5000);
		my_printer.f_message(PSTR(" heat: %i\n"),heat);
		delta = ((float)(heat -ref_heat)*PID_dT)/5.0 ;
		kf = (delta- target_temperature[selected_extruder]/4)/(float)fanSpeed;
		my_printer.f_message(PSTR(" f(100%%): %f \n"),kf);
	}
	else
	{
		my_printer.f_message(T(MSG_Temperature_Stabilisation_Abort_At),temp);
		//          switch off fan
		Set_FanSpeed(0);
		return false;
	}

	// now take the mean value

	kf += kf_50;
	kf += kf_75;
	kf /= 3.0;
	my_printer.f_message(PSTR("Kf = %f\n"),kf);

	if (update)
		config.Set_Kf(kf,selected_extruder);


	//          switch off fan
	Set_FanSpeed(0);
	return true;

}

